##########################
### complile quiz bank ###
##########################
# Format is:
# \begin{shinyQuizBank}{myBankName}
# \begin{shinyMCQ}{my multiple choice question text}
#   \shinyMCAR{my MC answer text}{my MC response text}
#   ...
# \end{shinyMCQ}
# ...
# \end{shinyQuizBank}
#
compileQuizBank = function(fname) {
  ### read .tex file and remove comments
  fname = checkFile(fname, ending=".tex")
  quizText = readLines(fname)
  comments = grep("^[%]", quizText)
  if (length(comments)>0) quizText[comments] = ""
  comments = regexpr("[^\\][%]", paste0(" ", quizText))
  Sel = (comments != -1)
  if (any(Sel)) quizText[Sel] = substring(quizText[Sel], 1, comments[Sel]-1)
  
  ### find \shinyQuizBank, verify structure, and get quiz bank name and options
  # Hide \{ and \} inside of $...$ or $$...$ using ctrl-a and ctrl-b
  quizText = gsub("[\\][{]", "\a", quizText) 
  quizText = gsub("[\\][}]", "\b", quizText) 
  #
  # Pull off \shinyQuizBank as --> full match | bank name | bank options
  FULL_MATCH = 1; BANK_NAME = 2; BANK_OPTIONS = 3
  quizBankB = str_match(quizText, "[\\\\]begin[[:blank:]]*[{]shinyQuizBank[}][[:blank:]]*[{](.*?)[}]([{].+?[}])?")
  quizBankBL = which(!is.na(quizBankB[, FULL_MATCH]))
  if (length(quizBankBL)==0) stop("no \\begin{shinyQuizBank}{bankName} in ", fname)
  if (length(quizBankBL)>1) stop("duplicate \\begin{shinyQuizBank}{bankName}s in ", fname)
  quizBankEL = grep("[\\]end[[:blank:]]*[{]shinyQuizBank[}]", quizText)
  if (length(quizBankEL)==0) stop("no \\end{shinyQuizBank}{bankName} in ", fname)
  if (length(quizBankEL)>1) stop("duplicate \\end{shinyQuizBank}{bankName}s in ", fname)
  if (quizBankBL>quizBankEL) stop("\\end{shinyQuizBank} before \\begin{shinyQuizBank}s in ", fname)
  #
  quizBankName = quizBankB[quizBankBL, BANK_NAME]
  if (quizBankName != make.names(quizBankName)) {
    stop("invalid quiz bank name '", quizBankName, 
         "'' (must start with a letter and have no blanks or punctuation)")
  }
  quizBankOptions = quizBankB[quizBankBL, BANK_OPTIONS]
  
  ### Locate the individual questions ###
  # Quiz counter to give each radioButton control a different name (across all banks)
  questionsSoFar = 0
  #
  # discard pre/postamble
  quizText = quizText[(quizBankBL+1):(quizBankEL-1)]
  lineNumOffset = quizBankBL
  #
  # find begin and end lines for each question
  # Parse \shinyMCQ{question text}{options} -> full match | {question text} | question text | {option} | option
  #
  # Allow \shinyMCQ{}[{fixed}] to span more than one line, while maintaining line numbers
  allMCQ = grep("[\\]begin[[:blank:]]*[{]shinyMCQ[}][[:blank:]]*[{]", quizText)
  allMCAR = grep("[\\]shinyMCAR", quizText)
  for (i in seq(along=allMCQ)) {
    line = allMCQ[i]
    if (i < length(allMCQ)) {
      endLine = allMCQ[i+1] - 1
    } else {
      endLine = length(quizText) - 1
    }
    possible = allMCAR[allMCAR>line & allMCAR<endLine]
    if (length(possible)>0) {
      endLine = possible[1] - 1
      if (endLine>line) {
        quizText[line] = paste(str_trim(quizText[line:endLine]), collapse="\n")
        quizText[line] = gsub("[\n]{2,}", "<p>", quizText[line])
        quizText[line] = gsub("[\n]", " ", quizText[line])
        quizText[(line+1):endLine] = ""
      }
    }
  }
  # 
  FULL_MATCH = 1; AFTER_TEXT = 2; QUESTION_TEXT = 4; QUESTION_OPTION = 6
  thisQuizB = str_match(quizText,
                        "[\\\\]begin[[:blank:]]*[{]shinyMCQ[}][[:blank:]]*(([{](.*?)[}])[[:blank:]]*([{](.*)[}])?)?")
  thisQuizE = grep("[\\]end[[:blank:]]*[{]shinyMCQ[}]", quizText)
  thisQuizBL = which(!is.na(thisQuizB[, FULL_MATCH]))
  noQ = is.na(thisQuizB[thisQuizBL, AFTER_TEXT])
  if (any(noQ)) {
    stop("@", thisQuizBL[which(noQ)[1]]+lineNumOffset, " \\shinyMCQ is missing its question")
  }
  thisQuizBE = c(thisQuizBL, thisQuizE)
  if (length(thisQuizBE)==0) {
    stop("quiz bank ", quizBankName, " is empty")
  }
  if (length(thisQuizBL) != length(thisQuizE)) {
    stop('@', paste(sort(c(thisQuizE,thisQuizBL))+lineNumOffset, collapse='/'), 
         ' un-balanced begin/end shinyMCQ statements')
  }
  # record state of "fixed" option for all questions
  questionOptions = thisQuizB[thisQuizBL, QUESTION_OPTION]
  questionOptions[is.na(questionOptions)] = ""
  questionOptions = str_trim(questionOptions)
  fixed = (regexpr("f(i(x(e(d?)?)?)?)?$", questionOptions, ignore.case=TRUE) != -1)
  #
  # make matrix of begin and end lines per question
  BEL = cbind(thisQuizBL, thisQuizE)
  
  # Prevent intertwined questions
  nMC = nrow(BEL)
  if (nMC>1) {
    temp = BEL[2:nMC,1] < BEL[1:(nMC-1),2]
    if (any(temp)) {
      temp = which(temp)[1]
      stop('@', paste(BEL[temp+1,1]+lineNumOffset, 
                      BEL[temp,2]+lineNumOffset, sep='/'), 
           ' intertwined shinyMCQ statements')
    }
  }
  
  # Get all question texts for this quiz bank
  thisQuizQuestionText = thisQuizB[thisQuizBL,QUESTION_TEXT]
  
  # Warnings for misuse
  temp = (questionOptions != "" & !fixed)
  if (any(temp)) {
    warning(thisQuizB[thisQuizBL[temp][1]], " (the only option allowed is 'fixed'")
  }
  #
  temp = unlist(lapply(1:nrow(BEL), function(x) BEL[x,1]:BEL[x,2]))
  temp = setdiff(1:length(quizText), temp)
  if (length(temp)>0) {
    tempT = quizText[temp]
    tempP = intersect(grep("^[[:blank:]]*$", tempT, invert=TRUE),
                      grep("^[[:blank:]]*[%]", tempT, invert=TRUE))
    if (length(tempP) > 0) {
      stop("@", temp[tempP]+lineNumOffset,
           " extraneous text found between questions for test bank '", quizBankName, "':",
           "'", tempT[tempP][1] , "'")
    }
  }
 
  # get quiz answers and responses for all questions into a list
  # (includes extensive detailed feedback on malformed questions)
  thisQuizResponses = lapply(1:nMC,
                             function(index) {
                               qErrTxt = thisQuizQuestionText[index]
                               if (nchar(qErrTxt)>20) qErrTxt = paste(substring(qErrTxt, 1, 20), "...")
                               at = paste0("@", BEL[index,1]+lineNumOffset, "-", BEL[index,2]+lineNumOffset)
                               if (BEL[index,2] == BEL[index,1]+1) {
                                 warning(at, ' question ', index, ' (', qErrTxt, ') of quiz bank ', '"', quizBankName,
                                         '" is empty')
                                 return(NULL)
                               }
                               text = quizText[(BEL[index,1]+1):(BEL[index,2]-1)]
                               text = text[grep("^[[:blank:]]*$", text, invert=TRUE)]
                               if (length(text)==0) {
                                 warning(at, ' question ', index, ' (', qErrTxt, ') of quiz bank ', '"', quizBankName,
                                         '" is empty')
                                 return(NULL)
                               }
                               sel = grep("[\\]shinyMCAR[*]?", text)
                               if (length(sel) == 0) {
                                 warning(at, ' question ', index, ' (', qErrTxt, ') of quiz bank ', '"', quizBankName,
                                         '" has no responses')
                                 return(NULL)
                               }
                               if (sel[1] != 1) {
                                 stop(at, ' question ', index, ' (', qErrTxt, ') of quiz bank ', '"', quizBankName,
                                      '" does not start with \\shinyMCAR')
                               }
                               if (length(sel) == 1) {
                                 selEnd = length(text)
                               } else {
                                 selEnd = c(sel[2:length(sel)]-1, length(text))
                               }
                               return(t(sapply(1:length(sel),
                                               function(ind2) {
                                                 one = paste(text[sel[ind2]:selEnd[ind2]], collapse="")
                                                 m = str_match(one, 
                                                               "[{](.*?)[}][[:blank:]]*[{](.*?)[}][[:blank:]]*([^%]*)?")[1,]
                                                 if (is.na(m[1])) {
                                                   stop(at, " question ", index, ' (', qErrTxt, ") of quiz bank '", quizBankName,
                                                        "' has malformed \\shinyMCAR #", ind2)
                                                 }
                                                 if (m[4]!="") {
                                                   stop(at, " invalid text in question ", index, ' (', qErrTxt, ") of quiz bank '",
                                                        quizBankName, "' after \\shinyMCAR #", ind2, "': '", str_trim(m[4]), "'")
                                                 }
                                                 m[2] = str_trim(m[2])
                                                 if (m[2]=="") {
                                                   stop(at, " empty answer in question ", index, ' (', qErrTxt, ") of quiz bank '",
                                                        quizBankName, "' in \\shinyMCAR #", ind2)
                                                 }
                                                 m[3] = str_trim(m[3])
                                                 if (m[3]=="") {
                                                   stop(at, " empty response in question ", index, ' (', qErrTxt, ") of quiz bank '", 
                                                        quizBankName, "' in \\shinyMCAR #", ind2)
                                                 }
                                                 if (m[4]!="") {
                                                   stop(at, " invalid text in question ", index, ' (', qErrTxt, ") of quiz bank '", 
                                                        quizBankName, "' after \\shinyMCAR #", ind2, "': '", str_trim(m[4]), "'")
                                                 }
                                                 return(m[2:3])
                                               })))
                             })
  assure(all(!is.null(thisQuizResponses)), 33991)
  nQuestions = length(thisQuizResponses)
  
  # store all quiz question Shiny HTML in a list named 'thisQuiz' in fooQuiz.RData (one for each quiz bank)
  thisQuiz = vector("list", nQuestions)
  #
  thisQuizQuestionText = doMathJax(thisQuizQuestionText, BEL[,"thisQuizBL"]+lineNumOffset, undoBraceHide=TRUE)
  locs = lapply(1:nMC,
                function(index) {
                  text = quizText[(BEL[index,1]+1):(BEL[index,2]-1)]
                    return(grep("[\\]shinyMCAR[*]?", text) + BEL[index,1] + lineNumOffset)
                  })
  for (i in 1:nQuestions) {
    questionsSoFar = questionsSoFar + 1
    if (is.null(thisQuizResponses[[i]])) thisQuizResponses[[i]] = cbind("*NO ANSWERS PROVIDED*","*NO RESPONSES PROVIDED*","")
    thisQuiz[[i]] = quizQuestion(quizBankName, thisQuizQuestionText[i], 
                                 questionChoices = doMathJax(thisQuizResponses[[i]][,1], locs[[i]], undoBraceHide=TRUE), 
                                 explanations = as.list(doMathJax(thisQuizResponses[[i]][,2], locs[[i]], undoBraceHide=TRUE)),
                                 questionsSoFar)
  }
  
  # Store all results in a binary "save" file
  attr(thisQuiz, "fixed") = fixed
  fname = paste0(quizBankName, "Quiz.RData")
  save(thisQuiz, file=fname)
  cat(fname, "successfully created.\n")
  invisible(NULL)
} # end compileQuizBank()
