# shinyTex auxilliary code  (required by shinyTex())
# H. Seltman, June 2015

if (!require("stringr")) stop("Please install R package 'stringr' to continue.")

###################################################
#### Quiz generator function for one quiz bank ####
###################################################
# Turn quizQuestion() output into a complete quiz
#  Arguments: 
#   bank: the name of the quiz bank (HTML is stored in "foo.RData" as list "thisQuiz")
#   addNumbers: add question numbers as "1) ", etc.
#   randomizeQuestions: present questions in a random order?
#   randomizeAnswers: present answers/reponses in a random order?
#   interElement: an HTML tag, eg. hr() or p() to go between questions
#  Value: a shiny.tag.list representing a quiz
#
quiz = function(bank, addNumbers=FALSE, randomizeQuestions=TRUE, randomizeAnswers=TRUE, max=NULL,
                interElement=NULL) {
  if (!require("shiny")) stop("Please install the 'shiny' package and try again.")
  
  # Get stored HTML from R save() file as R variable "thisQuiz" from <bank>Quiz.RData
  if (exists("thisQuiz")) rm(thisQuiz)
  fname = paste0(bank, "Quiz.RData")
  if (!file.exists(fname)) {  # file does not exists -- try to create it from likely source
    defaultBankFile = paste0(bank, ".tex")
    if (exists("compileQuizBank") && file.exists(defaultBankFile)) {
      warning("Note: compiling ", defaultBankFile)
      temp = try(compileQuizBank(defaultBankFile))
      if (is(temp, "try-error")) {
        warning("cannot compile quiz bank file '", defaultBankFile, "'")
        return(HTML(paste0("<p><strong>Quiz '", bank, "' goes here but '", fname, "' is unavailable.</strong></p>")))
      }
    } else {
      warning("cannot find file '", fname, "'")
      return(HTML(paste0("<p><strong>Quiz '", bank, "' goes here but '", fname, "' is unavailable.</strong></p>")))
    } # if trying to create non-existing quiz bank by compiling source
  } else { # file exists, but check if likely source is newer
    fTime = file.info(fname)$mtime
    sourceName = paste0(bank, ".tex")
    if (file.exists(sourceName) && file.info(sourceName)$mtime>fTime) {
      warning("Note: re-compiling modified ", sourceName)
      temp = try(compileQuizBank(sourceName))
      if (is(temp, "try-error")) warning("cannot compile quiz bank file '", sourceName, "'")
    }
  } # end if compiled quiz bank file exists
  err = try(load(fname))
  if (is(err, "try-error")) stop(fname, " is not a valid R saved object")
  
  # validate R save file value
  if (!exists("thisQuiz")) stop(fname, " did not contain 'thisQuiz'")
  nQ = length(thisQuiz)
  if (nQ==0) {
    warning("no valid questions for quiz bank '", bank, "'")
    return(HTML(paste0("<p><strong>Quiz '", bank, "' goes here but it has no questions.</strong></p>")))
  }
  isTag = sapply(thisQuiz, function(x) is(x, "shiny.tag"))
  assure(all(isTag), 77480)
  
  # randomize stored questions if needed
  fixed = attr(thisQuiz, "fixed") # reordering drops the attribute, so store it now
  if (randomizeQuestions) {
    qOrder = sample(nQ)
    thisQuiz = thisQuiz[qOrder]
  } else {
    qOrder = 1:nQ
  }
  
  # limit total number of questions
  Sel = 1:nQ
  if (is.numeric(max) && nQ>0 && nQ>max) {
    Sel = sort(sample(Sel, max))
    thisQuiz = thisQuiz[Sel]
    nQ = max
  }
  
  # setup answer randomizations
  if (randomizeAnswers == FALSE) {
    fixed = rep(FALSE, nQ)
  } else {
    if (is.null(fixed)) {
      fixed = rep(TRUE, nQ)
    } else {
      fixed = fixed[qOrder[Sel]]
    }
  }
    
  # Set characteristics of shiny.tag.list (questions, answers, and responses) generated by quizQuestion().
  # These are needed for randomization of answers and adding question numbers
  Q.AND.A=1; RESPONSE=2; Q.POS=1; A.POS1=2; A.POS2=2; A.POS3=1
  #
  # reorder answer/response pairs for each question
  if (randomizeAnswers) {
    for (qNum in 1:nQ) {
      if (!fixed[qNum]) {
        one = thisQuiz[[qNum]]$children[[Q.AND.A]]$children[[A.POS1]]$children[[A.POS2]]$children[[A.POS3]]
        nAns = length(one)
        ord = sample(nAns)
        thisQuiz[[qNum]]$children[[Q.AND.A]]$children[[A.POS1]]$children[[A.POS2]]$children[[A.POS3]] = 
          one[ord]
        thisQuiz[[qNum]]$children[[RESPONSE]]$children = thisQuiz[[qNum]]$children[[RESPONSE]]$children[ord]
      }
    }
  }
  
  # add numbering to questions within a bank  (done here due to possible randomization)
  if (addNumbers) {
    for (qNum in 1:nQ) {
      temp = thisQuiz[[qNum]]$children[[Q.AND.A]]$children[[Q.POS]]
      if (class(temp) == "shiny.tag.list") {
        assure(length(temp)>1 && !is.null(temp[[1]]$name) && temp[[1]]$name == "p", 924075)
        assure(!is.null(temp[[1]]$children) && length(temp[[1]]$children)>0 && 
               is.character(temp[[1]]$children[[1]]), 734679)
        temp[[1]]$children[[1]] = paste(qNum, ") ", temp[[1]]$children[[1]], sep="")
      } else {
        assure(class(temp) == "character", 475610)
        temp = paste(qNum, ") ", temp, sep="")
      }
      thisQuiz[[qNum]]$children[[Q.AND.A]]$children[[Q.POS]] = temp
    }
  }
  
  # add inter-question element
  if (nQ>1 && !is.null(interElement)) {
    temp = thisQuiz
    thisQuiz = vector("list", 2*nQ-1)
    thisQuiz[seq(1,2*nQ-1,2)] = temp
    for (i in seq(2,2*nQ-2,2)) thisQuiz[[i]] = interElement
  }
  
  lstOut = do.call(list, thisQuiz)
  class(lstOut) = c("shiny.tag.list", "list")
  return(lstOut)
}



# Similar to c function assure(), this function quits with the specified message if
# a condition is false (or goes to the browser if file "debug.txt" exists)
#
assure = function(shouldBeTrue, errorNum) {
  if (!shouldBeTrue) {
    if (file.exists("debug.txt")) {
      browser()
    } else {
      stop("programmer error ", errorNum)
    }
  }
  invisible(NULL)
}
